#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <inttypes.h>
#include <errno.h>
#include <iostream.h>
#include <pthread.h>
#include <signal.h>
#include <sys/neutrino.h>
#include <sys/syspage.h> 

static const int TEMP = 500;
static int numclient = 1;

iostream& operator <<( iostream& c, char* s ) {
   static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
   pthread_mutex_lock( &mutex );
   c << s << flush;
   pthread_mutex_unlock( &mutex );
   return c;
};

static uint64_t tb; 
inline uint64_t mark( void ) { 
   const static uint64_t cps = SYSPAGE_ENTRY( qtime )->cycles_per_sec;  // частота процессора:
   return ( ClockCycles() - tb ) * 1000 / cps;
};

const int MSGLEN = 80;

void* server ( void* chid ) {
   int rcvid;
   char message[ MSGLEN ];   
   while( true ) {
      rcvid = MsgReceive( (int)chid, message, MSGLEN, NULL );
      sprintf( message + strlen( message ), "[%07llu] ... ", mark() );
      delay( TEMP );
      sprintf( message + strlen( message ), "[%07llu]->", mark() );      
      MsgReply( rcvid, EOK, message, strlen( message ) + 1 );
   };
   return NULL;
};

void* client ( void* data ) {
   while( true ) {
      char message[ MSGLEN ];   
      sprintf( message, "%d:\t[%07llu]->", pthread_self(), mark() ); 
      MsgSend( (int)data, message, strlen( message ) + 1, message, MSGLEN );
      sprintf( message + strlen( message ), "[%07llu]", mark() );
      cout << message << endl; 
      static unsigned int seed = 0;
      delay( numclient * ( ( (long)rand_r( &seed ) * TEMP / RAND_MAX ) + TEMP ) );
   };
   return NULL;
};                    

int main( int argc, char **argv ) {
   if( argc > 1 && atoi( argv[ 1 ] ) > 0 ) numclient = atoi( argv[ 1 ] );
   tb = ClockCycles();   
   int chid = ChannelCreate( 0 );
   if( pthread_create( NULL, NULL, server, (void*)chid ) != EOK )
      perror( "server create" ), exit( EXIT_FAILURE );
   for( int i = 0; i < numclient; i++ ) 
      if( pthread_create( NULL, NULL, client,
             (void*)ConnectAttach( 0, 0, chid, _NTO_SIDE_CHANNEL, 0 ) ) != EOK )
         perror( "client create" ), exit( EXIT_FAILURE );
   sigpause( SIGINT );
   return EXIT_SUCCESS;
};

