#include <errno.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>
#include <sys/iofunc.h>
#include <sys/dispatch.h>

// Чтение открытого канала
static int prior_read( resmgr_context_t *ctp, io_read_t *msg, RESMGR_OCB_T *ocb ) {
   static bool odd = true;
   int status = iofunc_read_verify( ctp, msg, ocb, NULL );
   if( status != EOK ) return status;
   if( msg->i.xtype & _IO_XTYPE_MASK != _IO_XTYPE_NONE ) return ENOSYS;    
   if( odd ) { 
      struct sched_param param;
      sched_getparam( 0, &param );
      static char rbuf[ 4 ];
      sprintf( rbuf, "%d\n", param.sched_curpriority );
      MsgReply( ctp->rcvid, strlen( rbuf ) + 1, rbuf, strlen( rbuf ) + 1 );
   }
   else  MsgReply( ctp->rcvid, EOK, NULL, 0 );
   odd = !odd;   
   return  _RESMGR_NOREPLY;
};    

main( int argc, char **argv ) {
    resmgr_attr_t        resmgr_attr;
    dispatch_t           *dpp;
    dispatch_context_t   *ctp;
    int                  id;
    /* initialize dispatch interface */
    if( ( dpp = dispatch_create() ) == NULL ) {
        fprintf( stderr, "%s: Unable to allocate dispatch handle.\n", argv[0] );
        return EXIT_FAILURE;
    }
    /* initialize resource manager attributes */
    memset( &resmgr_attr, 0, sizeof resmgr_attr );
    resmgr_attr.nparts_max = 1;
    resmgr_attr.msg_max_size = 2048;
    /* initialize functions for handling messages */
   static resmgr_connect_funcs_t  connect_funcs;
   static resmgr_io_funcs_t  io_funcs;
    iofunc_func_init( _RESMGR_CONNECT_NFUNCS, &connect_funcs, _RESMGR_IO_NFUNCS, &io_funcs );
    // прописываем свои функции в таблицы операций:
    io_funcs.read = prior_read;
    /* initialize attribute structure used by the device */
    static iofunc_attr_t  attr;    
    iofunc_attr_init( &attr, S_IFNAM | 0666, 0, 0 );
    /* attach our device name */
    id = resmgr_attach( dpp,                    /* dispatch handle        */
                                    &resmgr_attr,      /* resource manager attrs */
                                    "/dev/prior", 
                                    _FTYPE_ANY,        /* open type              */
                                    0,                        /* flags                  */
                                    &connect_funcs, /* connect routines       */
                                    &io_funcs,           /* I/O routines           */
                                    &attr );                /* handle                 */
    if( id == -1 ) {
        fprintf( stderr, "%s: Unable to attach name.\n", argv[ 0 ] );
        return EXIT_FAILURE;
    }
    /* allocate a context structure */
    ctp = dispatch_context_alloc( dpp );
    /* start the resource manager message loop */
    while( true ) {
        if( ( ctp = dispatch_block( ctp ) ) == NULL ) {
            fprintf( stderr, "block error\n" );
            return EXIT_FAILURE;
        }
        dispatch_handler( ctp );
    }
};

