#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <limits.h>
#include <dirent.h> 
#include <iostream.h>
#include <queue>

//---------------------------------------------------------------------------------------------------------------------------
// Переопределение THREAD_POOL_PARAM_T таким образом, чтобы использование
// функций dispatch_*() не вызывало предупреждений компиляции
#define THREAD_POOL_PARAM_T dispatch_context_t
#include <sys/dispatch.h>
//---------------------------------------------------------------------------------------------------------------------------
// Все последующие описания до реализации функций обработки операций (open, read, write ...)
// делается для того, чтобы переопределить (добавить ссылку на экземпляр Client) стандартную
// структуру ocb перед включением <sys/iofunc.h>. Это нужно для того, чтобы каждый открытый
// в клиентской программе файловый дескриптор ссылался на отдельный GUI-компонент.
#define IOFUNC_OCB_T struct ownocb
#include <sys/iofunc.h>

class ownocb : public iofunc_ocb_t  {
   static const int BUFSIZE = 1024;
public: 
   char *buf;
   ownocb( void ) { buf = new char[  BUFSIZE ]; };
   ~ownocb( void ) { delete buf; }; 
};

IOFUNC_OCB_T *ownocb_calloc( resmgr_context_t *ctp, IOFUNC_ATTR_T *device ) { return new ownocb; };
void ownocb_free( IOFUNC_OCB_T *o ) { delete o; };
iofunc_funcs_t ownocb_funcs = { _IOFUNC_NFUNCS, ownocb_calloc, ownocb_free };
iofunc_mount_t mountpoint = { 0, 0, 0, 0, &ownocb_funcs };

static int nolock( resmgr_context_t *ctp, void *v, IOFUNC_OCB_T *ocb ) { return EOK; };

// Чтение открытого канала
static int line_read( resmgr_context_t *ctp, io_read_t *msg, IOFUNC_OCB_T *ocb ) {
   if( strlen( ocb->buf ) != 0 ) {
      MsgReply( ctp->rcvid, strlen( ocb->buf ) + 1, ocb->buf, strlen( ocb->buf ) + 1 );
      strcpy( ocb->buf, "" );
   }
   else MsgReply( ctp->rcvid, EOK, NULL, 0 );
   return  _RESMGR_NOREPLY;
};    

// Запись в открытый канал
static int line_write( resmgr_context_t *ctp, io_write_t *msg, IOFUNC_OCB_T *ocb ) {
   resmgr_msgread( ctp, ocb->buf, msg->i.nbytes, sizeof( msg->i ) );
   _IO_SET_WRITE_NBYTES( ctp, msg->i.nbytes );
   return  EOK;        
};

const char sResName[ _POSIX_PATH_MAX + 1 ] = "/dev/wmng";

// Запуск ресурс-менеджера отдельным потоком
static void StartResMng( void ) {
   dispatch_t*  dpp;
   if( ( dpp = dispatch_create() ) == NULL ) perror( "dispatch create" ), exit( EXIT_FAILURE );
   resmgr_attr_t  resmgr_attr;
   memset( &resmgr_attr, 0, sizeof resmgr_attr );
   resmgr_attr.nparts_max = 1;
   resmgr_attr.msg_max_size = 2048;
   // статичность этих определений принципиально важна
   static resmgr_connect_funcs_t  connect_funcs;
   static resmgr_io_funcs_t  io_funcs;
   static iofunc_attr_t  attr;
   iofunc_func_init( _RESMGR_CONNECT_NFUNCS, &connect_funcs, _RESMGR_IO_NFUNCS, &io_funcs );
   // прописываем свои функции в таблицы операций (open & close не прописаны, 
   // поскольку испьзуются втроенные реализации по умолчанию):
   io_funcs.read = line_read;
   io_funcs.write = line_write;
   io_funcs.lock_ocb = nolock;    
   iofunc_attr_init( &attr, S_IFNAM | 0666, NULL, NULL );
   // переопределены функции создания-уничтожения ocb
   attr.mount = &mountpoint;
   // присоединить имя нового ресурса
   if( resmgr_attach( dpp, &resmgr_attr, sResName, _FTYPE_ANY,
                                0, &connect_funcs, &io_funcs, &attr ) == -1 )
      perror( "name attach" ), exit( EXIT_FAILURE );
   // инициализация пула потоков многопотокового менеджера
   thread_pool_attr_t  pool_attr;
   memset( &pool_attr, 0, sizeof pool_attr );
   pool_attr.handle = dpp;
   pool_attr.context_alloc = dispatch_context_alloc;
   pool_attr.block_func = dispatch_block;
   pool_attr.handler_func = dispatch_handler;
   pool_attr.context_free = dispatch_context_free;
   pool_attr.lo_water = 2;
   pool_attr.hi_water = 6;    
   pool_attr.increment = 1;
   pool_attr.maximum = 50;
   // создание пула потоков и его запуск
   thread_pool_t*  tpp;
   if( ( tpp = thread_pool_create( &pool_attr, POOL_FLAG_EXIT_SELF ) ) == NULL ) 
      perror( "pool create" ), exit( EXIT_FAILURE );
   thread_pool_start( tpp );
};

// Старт приложения (инициализации ещё до библиотек Poton)
int main( int argc, char *argv[] ) {
   // проверка присутствия ранее загруженноо экземпляра ресурс-менеджера, 2 экз. нам ни к чему!
   char sDirName[ _POSIX_NAME_MAX + 1 ];
   int nDirLen = strrchr( (const char*)sResName, '/' ) - (char*)sResName;
   strncpy( sDirName, sResName, nDirLen );
   sDirName[ nDirLen ] = '\0';
   DIR *dirp = opendir( sDirName ); 
   if( dirp == NULL ) perror( "directory not found" ), exit( EXIT_FAILURE );
   struct dirent *direntp; 
   while( true ) {
      if( ( direntp = readdir( dirp ) ) == NULL ) break;
      if( strcmp( direntp->d_name, strrchr( sResName, '/' ) + 1 ) ==  0 ) 
         cout << "second copy of manager" << endl, exit( EXIT_FAILURE );
   }; 
   closedir( dirp );   
   StartResMng();    
   exit( EXIT_SUCCESS );
};

