#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#include <signal.h>
#include <unistd.h>
#include <sched.h>

static void handler( int signo, siginfo_t* info, void* context ) {
   int oldprio = getprio( 0 );
   setprio( 0, info->si_value.sival_int );
   cout << "SIG = " << signo << ": old priority = " << oldprio
           << ", new priority = " << getprio( 0 ) << endl;
   setprio( 0, oldprio );       
};

int main( int argc, char *argv[] ) {
   // установка обработчика сигнала
   sigset_t sig;
   sigemptyset( &sig ); 
   sigaddset( &sig, SIGUSR1 );                  // #define SIGUSR1     16 
   sigprocmask( SIG_BLOCK, &sig, NULL );     
   struct sigaction act;
   act.sa_mask = sig;
   act.sa_sigaction = handler;
   act.sa_flags = SA_SIGINFO;
   if( sigaction( SIGUSR1 , &act, NULL ) < 0 ) perror( "set signal handler" ), exit( EXIT_FAILURE );
   // запуск синхронизирующего процесса
   const char *prgname = "./p1ch", *sdelay = "3";
   pid_t pid =
      ( ( argc > 1 ) && ( atoi( argv[ 1 ] ) >= sched_get_priority_min( SCHED_RR ) ) &&
        ( atoi( argv[ 1 ] ) <= sched_get_priority_max( SCHED_RR ) ) ) ?
      spawnl( P_NOWAIT, prgname, prgname, sdelay, argv[ 1 ], NULL ) :
      spawnl( P_NOWAIT, prgname, prgname, sdelay, NULL );
   if( pid == -1 ) perror( "spawn child process" ), exit( EXIT_FAILURE );
   // разблокироваться для приёма сигнала
   sigprocmask( SIG_UNBLOCK, &sig, NULL );        
   while( true ) {
      if( sleep( 3 ) != 0 ) continue;
      cout << "parent main loop: priority = " << getprio( 0 ) << endl;
   }; 
};

