#include "common.h"
#include <sys/resmgr.h>
#include <sys/dispatch.h>
#include <sys/iofunc.h>

result data;
//-----------------------------------------------------------------------------------------------------
// чтение блока результата по read()
static int readfunc( resmgr_context_t *ctp, io_read_t *msg, iofunc_ocb_t *ocb ) {
   int sts = iofunc_read_verify( ctp, msg, ocb, NULL );
   if( sts != EOK ) return sts;
   // ... и сразу же - ответ:
   MsgReply( ctp->rcvid, sizeof( result ), &data, sizeof( result ) );
   // менеджер отвечать не будет - мы уже ответили
   return  _RESMGR_NOREPLY;
}; 
//---------------------------------------------------------------------------------------------------------------------------
// Управление открым каналом (devctl):
static int devctlfunc( resmgr_context_t *ctp, io_devctl_t *msg, iofunc_ocb_t *ocb ) { 
   int sts = iofunc_devctl_default( ctp, msg, ocb );
   if( sts != _RESMGR_DEFAULT ) return sts;
   // здесь должна бы быть проверка кода команды dvctl(), но нас интересует только длина данных...
   unsigned int nbytes = ( msg->i.dcmd -  DCMD_SRR ) >> 16;
   msg->o.nbytes = nbytes;
//cout << nbytes << " " << flush;
   return _RESMGR_PTR( ctp, &msg->i, sizeof( msg->i ) + nbytes  );
};
//-----------------------------------------------------------------------------------------------------
// запуск ресурс-менеджера отдельным потоком
static void* install( void* data ) {
   // старт однопотокового менеджера
   dispatch_t  *dpp;
   if( ( dpp = dispatch_create() ) == NULL ) exit( "dispatch allocate" );
   resmgr_attr_t  resmgr_attr;
    //инициализация аттрибутов ресурс менеджера   
   memset( &resmgr_attr, 0, sizeof( resmgr_attr ) );
   resmgr_attr.nparts_max = 1;
   resmgr_attr.msg_max_size = 2048;
    // таблицы функций подсоединения и обмена
   static resmgr_connect_funcs_t   connect_funcs;
   static resmgr_io_funcs_t             io_funcs;
   iofunc_func_init( _RESMGR_CONNECT_NFUNCS, &connect_funcs,
                             _RESMGR_IO_NFUNCS, &io_funcs );
   // прописываем свои функции в таблицы операций 
   io_funcs.read = &readfunc;
   io_funcs.devctl = &devctlfunc;
   static iofunc_attr_t  attr;
   // инициализация структуры аттрибутов, используемых устройством
   iofunc_attr_init( &attr, S_IFNAM | 0666, 0, 0 );
   // присоединить имя нового ресурса
   if( resmgr_attach( dpp, &resmgr_attr,	DEVNAME, _FTYPE_ANY,
                              0, &connect_funcs, &io_funcs, &attr ) == -1 ) exit( "prefix attach" );
   // разместить структуру контекста
   dispatch_context_t  *ctp = dispatch_context_alloc( dpp );
   // старт бескоречного цикла, его завершение может быть 
   // обеспечено только извне командами kill или slay
   while( true ) {
      if( ( ctp = dispatch_block( ctp ) ) == NULL ) exit( "block error" );
      dispatch_handler( ctp );
   };
};
//-----------------------------------------------------------------------------------------------------
const int blk = 100000;
// функция потока канала ввода/вывода сообщений
void* msginout( void* c ) {
   static uint8_t bufin[ blk ];
   struct _msg_info info;
   while( true ) {
      int rcvid = MsgReceive( data.chid, &bufin, blk, &info );
// cout << info.msglen << " ";
      if( rcvid < 0 ) exit( "message receive" );
      if( MsgReply( rcvid, EOK, &bufin, info.msglen ) < 0 ) exit( "message reply" );
   };
};
//-----------------------------------------------------------------------------------------------------
// "пустой" обработчик сигнала (SIGINT - ^C)
inline static void empty( int signo ) { };
//-----------------------------------------------------------------------------------------------------
int main( int argc, char *argv[] ) {
   cout << "SRR server: " << VERSION << endl;
   // проверить наличие запущенной копии - 2 экземпляра не запускать!
   int fd = open( DEVNAME, O_RDONLY );
   if( fd > 0 ) close( fd ), cout << "already in use: " << DEVNAME << endl, exit( EXIT_FAILURE ); 
   cout << ". . . . . . waiting  ^C. . . . . ." << flush;
    // переопределить (запретить) ^C
   signal( SIGINT, empty ); 
   // создать канал низкоуровневого обмена
   data.chid = ChannelCreate( 0 );
   // установить приёмо-передатчик сообщений
   if( pthread_create( NULL, NULL, msginout, NULL ) != EOK ) exit( "mesage thread" );
   // установить менеджер обмена сообщениями отдельным потоком
   if( pthread_create( NULL, NULL, install, NULL ) != EOK ) exit( "menager thread" );
// cout << endl;
   pause();
   cout << "Finalisation...                                                        " << endl;
   ChannelDestroy( data.chid );
   return EXIT_SUCCESS;
};

