#include <stdlib.h>
#include <iostream.h>
#include <sys/syspage.h> 
#include <inttypes.h>
#include <unistd.h>
#include <sys/neutrino.h>
#include <pthread.h>
#include <list>

typedef int element;

class dbase : public list<element> {
   static const int READ_DELAY = 1, WRITE_DELAY = 2;
   pthread_rwlock_t loc;   
public: 
   dbase( void ) { pthread_rwlock_init( &loc, NULL ); };
   ~dbase( void ) { pthread_rwlock_destroy( &loc ); };
   void add( const element& e ) {
      pthread_rwlock_wrlock( &loc );
      int pos = size() * rand() / RAND_MAX;
      list<element>::iterator p = begin();
      for( int i = 0; i < pos; i++ ) p++;
      insert( p, e );
      delay( WRITE_DELAY );
      pthread_rwlock_unlock( &loc );
   };
   int pos( const element& e ) {
      int n = 0;
      pthread_rwlock_rdlock( &loc );      
      for( list<element>::iterator i = begin(); i != end(); i++, n++ ) 
         if( *i == e ) { delay( READ_DELAY ); break; };
      pthread_rwlock_unlock( &loc );
      if( n == size() )  n = -1;
      return n;
   };
} data;

inline element erand( unsigned long n ) { return (element)( ( n * rand() ) / RAND_MAX ); };
inline bool wrand( double p ) { return (double)rand() / (double)RAND_MAX < p; };

static void* add( void* par ) { data.add( (element)par ); };

static void* pos( void* par ) { data.pos( (element)par ); };

int main( int argc, char *argv[] ) {
   static unsigned long n = 1000;           // число обращений к базе данных
   static double p = .1;                            // частота операций обновления базы данных
   unsigned long m;
   if( argc > 1 && ( m = atoll( argv[ 1 ] ) ) != 0 ) n = m;
   double q;
   if( argc > 2 && ( q = atof( argv[ 2 ] ) ) != 0 ) p = q;
   for( int i = 0; i < n; i++ ) data.add( erand( n ) );
   const uint64_t cps = SYSPAGE_ENTRY( qtime )->cycles_per_sec;  // частота процессора:
   pthread_t *h = new pthread_t[ n ];
   uint64_t t = ClockCycles();
   for( int i = 0; i < n; i++ ) {
      element e = erand( n );
      pthread_create( h + i, NULL, wrand( p ) ? add : pos, (void*)e );
   };
   for( int i = 0; i < n; i++ ) pthread_join( h[ i ], NULL );
   t = ( ( ClockCycles() - t ) * 1000000000 ) / cps;
   delete h;
   cout << "evaluation time: " << (double)t / 1.E9 << " sec." << endl;
   return EXIT_SUCCESS;
};

