#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include <iostream.h>
#include <unistd.h>
#include <pthread.h>
#include <errno.h>
#include <sys/neutrino.h>

unsigned long N = 1000;
static pthread_barrier_t bstart;
static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
static bool debug = false;
static char *str;
static volatile int ind = 0;

void* threadfunc ( void* data ) {
   pthread_barrier_wait( &bstart );
   unsigned long i = 0;
   char tid[ 8 ];
   sprintf( tid, "%d", pthread_self() ); 
   uint64_t t = 0, t1;
   while( i++ != N ) {
      t1 = ClockCycles();
      pthread_mutex_lock( &mutex );
      if( debug ) str[ ind++ ] = *tid;
      pthread_mutex_unlock( &mutex );  
      t += ClockCycles() - t1;
      sched_yield();
   };
   cout << pthread_self() << "\t: cycles - " << t << "; on mutex - " << t / N << endl;
   return NULL;
};

int main( int argc, char *argv[] ) {
   int opt, val; 
   while ( ( opt = getopt( argc, argv, "n:v" ) ) != -1 ) {
      switch( opt ) {
         case 'n' :
            if( sscanf( optarg, "%i", &val ) != 1 )
               cout << "parse command line error" << endl, exit( EXIT_FAILURE );
            if( val > 0 ) N = val;
            break;
         case 'v' :
            debug = true;
            break;
         default :
            cout << "parse command line failed" << endl, exit( EXIT_FAILURE );
            break;
       }
   };
   if( debug ) str = new char [ 2 * N + 1 ];   
   const int T = 2;
   pthread_t tid[ T ];
   if( pthread_barrier_init( &bstart, NULL, T ) != EOK ) perror( "barrier init" ), exit( EXIT_FAILURE );
   for( int i = 0; i < T; i++ )
      if( pthread_create( tid + i, NULL, threadfunc, NULL ) != EOK )
         perror( "thread create" ), exit( EXIT_FAILURE );
   for( int i = 0; i < T; i++ ) pthread_join( tid[ i ], NULL );
   if( debug ) {
      str[ ind ] = '\0';
      cout << str << endl;
      delete [] str;
   };
   exit( EXIT_SUCCESS );
};

