#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include <iostream.h>
#include <unistd.h>
#include <pthread.h>
#include <errno.h>
#include <sys/neutrino.h>
#include <semaphore.h>

unsigned long N = 1000;
unsigned int T = 2;

static sem_t* sem;
static bool debug = false;
static char *str;
static volatile int ind = 0;
uint64_t *t;

void* threadfunc ( void* data ) {
   unsigned long i = 0;
   char tid[ 8 ];
   sprintf( tid, "%X", pthread_self() ); 
   if( (int)data == T - 1 ) {
      uint64_t c =  ClockCycles();
     for( int i = 0; i < T; i++ ) t[ i ] = c;
   };
   while( i++ < N ) {
      sem_wait( sem + (int)data );
      if( debug ) str[ ind++ ] = *tid;
      sem_post( sem + ( (int)data + 1 ) % T );
   };
   t[ (int)data ] = ClockCycles() - t[ (int)data ];
   return NULL;
};

int main( int argc, char *argv[] ) {
   int opt, val; 
   while ( ( opt = getopt( argc, argv, "n:t:v" ) ) != -1 ) {
      switch( opt ) {
         case 'n' :
            if( sscanf( optarg, "%i", &val ) != 1 )
               cout << "parse command line error" << endl, exit( EXIT_FAILURE );
            if( val > 0 ) N = val;
            break;
         case 't' :
            if( sscanf( optarg, "%i", &val ) != 1 )
               cout << "parse command line error" << endl, exit( EXIT_FAILURE );
            if( val > 0 ) T = val;
            break;
         case 'v' :
            debug = true;
            break;
         default :
            cout << "parse command line failed" << endl, exit( EXIT_FAILURE );
            break;
       }
   };
   if( debug ) str = new char [ T * N + 1 ];   
   pthread_t* tid = new pthread_t[ T ];
   sem = new sem_t[ T ];   
   t = new uint64_t[ T ];
   for( int i = 0; i < T; i++ ) {
      if( sem_init( sem + i, 0, ( i == ( T - 1 ) ) ? 1 : 0 ) )
         perror( "semaphore init" ), exit( EXIT_FAILURE );
      if( pthread_create( tid + i, NULL, threadfunc, (void*)i ) != EOK )
         perror( "thread create error" ), exit( EXIT_FAILURE );
   };
   for( int i = 0; i < T; i++ ) pthread_join( tid[ i ], NULL );
   for( int i = 0; i < T; i++ ) sem_destroy( sem + i );
   delete [] sem;
   for( int i = 0; i < T; i++ ) 
      cout << tid[ i ] << "\t: cycles - " << t[ i ] << ";\ton semaphore - " << t[ i ] / T / N << endl;
   delete [] tid;
   delete [] t;
   if( debug ) {
      str[ ind ] = '\0';
      cout << str << endl;
      delete [] str;
   };
   exit( EXIT_SUCCESS );
};

