#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#include <unistd.h>
#include <limits.h>
#include <pthread.h>
#include <inttypes.h>
#include <sys/neutrino.h>
#include <sys/syspage.h>
#include <errno.h>
#include <sched.h>
//#include <math.h>

// преобразование процессорных циклов в милисекунды:
static double cycle2milisec ( uint64_t ccl ) {
   const static double s2m = 1.E+3;
   const static uint64_t cps = SYSPAGE_ENTRY( qtime )->cycles_per_sec;  // частота процессора:
   return (double)ccl * s2m / (double)cps; 
};

static int nsingl = 1;   

// рабочая функция
void workproc( void ) {
   const int nrep = 1000, msingl = 30000;   
   for( int j = 0; j < nrep; j++ )  // ... якобы вычисления
      for( uint64_t i = 0; i < msingl * nsingl; i++ ) i = ( i + 1 ) - 1; 
}; 

static pthread_barrier_t bfinish;

uint64_t trtime;

void* threadfunc ( void* data ) {
   trtime = ClockCycles();
   workproc();
   trtime = ClockCycles() - trtime;
   pthread_barrier_wait( &bfinish );   
   return NULL;
};                    

int main( int argc, char *argv[] ) {
   int opt, val, polisy = SCHED_RR, nall = SHRT_MAX;
   while ( ( opt = getopt( argc, argv, "n:p:frs" ) ) != -1 ) {
      switch( opt ) {
         case 'p' :
            if( sscanf( optarg, "%i", &val ) != 1 ) perror( "parse command line failed" ), exit( EXIT_FAILURE );
            if( ( val != getprio( 0 ) ) &&
                ( val >= sched_get_priority_min( polisy ) ) &&
                ( val <= sched_get_priority_max( polisy ) ) ) 
               if( setprio( 0, val ) == -1 ) perror( "priority isn't a valid" ), exit( EXIT_FAILURE );
            break;
         case 'n' :
            if( sscanf( optarg, "%i", &val ) != 1 ) perror( "parse command line failed" ), exit( EXIT_FAILURE );
            if( val > 0 ) nsingl *= val;
            break;
         case 'f' : polisy = SCHED_FIFO; break;
         case 'r' : polisy = SCHED_RR; break;
         case 's' : polisy = SCHED_SPORADIC; break;         
         default :
            perror( "parse command line failed" ), exit( EXIT_FAILURE );
            break;
       }
   };
   timespec interval;
   sched_rr_get_interval( 0, &interval );
   cout << "Priority level: " << getprio( 0 ) 
           << " [" << sched_get_priority_min( polisy ) << "..."
           << sched_get_priority_max( polisy ) << "]" << endl  
           << "Rescheduling interv. = " << (double)interval. tv_nsec / 1000000. << " msec" << endl;
   if( pthread_barrier_init( &bfinish, NULL, 2 ) != EOK ) perror( "barrier init" ), exit( EXIT_FAILURE );      
   pthread_attr_t attr;
   pthread_attr_init( &attr );
   pthread_attr_setschedpolicy( &attr, polisy );
   pthread_attr_setinheritsched( &attr, PTHREAD_EXPLICIT_SCHED );
   if( pthread_create( NULL, &attr, threadfunc, NULL ) != EOK )
      perror( "thread create" ), exit( EXIT_FAILURE );
   pthread_attr_destroy( &attr );      
   pthread_barrier_wait( &bfinish );   
   cout << "Evaluation time = " << cycle2milisec ( trtime ) << " msec" << endl;
   pthread_barrier_destroy( &bfinish );   
   exit( EXIT_SUCCESS );      
};

